/* Set the backlight level on a toshiba TECRA S1.
   Copyright ( C ) 2005-2006 Thorsten Zachmann <t.zachmann@zagge.de>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or ( at your option ) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
Changelog:   

0.1.1
    Fix: Save and restore the eax when triggering SMI
    Add missing include for exit()
 */
#include <sys/io.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#define BUFFER_SIZE        0x20

#define SET_LCD_BRIGHTNESS 0xa2
#define GET_LCD_BRIGHTNESS 0xa3
#define SET_DISPLAY_STATE  0xa7
#define GET_DISPLAY_STATE  0xa8
#define SET_FNF5_INTERFACE 0xc2

#define SMI_PORT 0xb2


void smi_command( int function,
                  int inputbuffer[BUFFER_SIZE],
                  int inputsize,
                  int outputbuffer[BUFFER_SIZE] )
{
    int counter = 0;
    int a = 0;
    int input;
    int x;
    int y;
    int smi;

    if ( iopl(3) != 0 )
    {
        printf("iopl(3) failed. Please run as root.\n");
        exit(1);
    }

    do
    {
        a = counter;
        a += 0x60;

        outb( a, 0x72 );
        if ( counter < inputsize )
        {
            input = inputbuffer[counter];
        }
        else
        {
            input = 0;
        }
        outb( input, 0x73 );
        counter++;
    } while ( counter < BUFFER_SIZE );

    outl( 0x8000f840 , 0xcf8 );
    x = inb( 0xcfc );
    x = x & 0xff80;
    x += 0x2c;

    y = inl(x);

    outl( 0, x );

    smi = function;
    smi = smi << 8;
    smi |= 0xe4;

    {
        int retval = 0x0000;
        int trigger = ( function << 8 ) | 0x00e4;

        __asm__ __volatile__ ( 
                "push %%eax\n\t"        // save eax
                "out  %%ax, %%dx \n\t"  // trigger SMI. Upon success this clears eax
                "mov  %%eax, %%ecx\n\t" // store value of eax in ecx
                "pop  %%eax\n\t"        // restore eax
                : "=c" ( retval )
                : "a" ( trigger ), "d" ( SMI_PORT )
        );
    }


    outl( y, x );

    counter = 0;
    do
    {
        a = counter;
        a += 0x60;
        outb( a, 0x72 );
        input = inb( 0x73 );
        outputbuffer[counter] = input;
        counter++;
    } while ( counter < BUFFER_SIZE );

    iopl(0);
}

void set_brightness( int brightness )
{
    int inputbuffer[BUFFER_SIZE];
    int outputbuffer[BUFFER_SIZE];

    if ( brightness < 0 || brightness > 7 )
        return;

    inputbuffer[0] = brightness;
    smi_command( SET_LCD_BRIGHTNESS, inputbuffer, 1, outputbuffer );
}

int get_brightness()
{
    int inputbuffer[BUFFER_SIZE];
    int outputbuffer[BUFFER_SIZE];

    smi_command( GET_LCD_BRIGHTNESS, inputbuffer, 0, outputbuffer );
    return outputbuffer[0];
}

void set_display_state( int state )
{
    int inputbuffer[BUFFER_SIZE];
    int outputbuffer[BUFFER_SIZE];

    if ( state < 0 || state > 4 )
        return;

    inputbuffer[0] = state;
    smi_command( SET_DISPLAY_STATE, inputbuffer, 1, outputbuffer );
}

int get_display_state()
{
    int inputbuffer[BUFFER_SIZE];
    int outputbuffer[BUFFER_SIZE];

    smi_command( GET_DISPLAY_STATE, inputbuffer, 0, outputbuffer );
    return outputbuffer[0];
}

void set_fnf5_interface( int state )
{
    int inputbuffer[BUFFER_SIZE];
    int outputbuffer[BUFFER_SIZE];

    if ( state < 0 || state > 1 )
        return;

    inputbuffer[0] = state;
    smi_command( SET_FNF5_INTERFACE, inputbuffer, 1, outputbuffer );
}

void usage( char * programname )
{
    printf( "usage: %s [-h] [-s level] [-g]\n", programname );
    printf( "-g        get the LCD brightness level\n" );
    printf( "-h        print this message\n" );
    printf( "-s level  set the LCD brightness to level (level 0-7)\n" );
}

int main( int argc, char *argv[] )
{
    int op;
    int s_brightness = -1;
    int g_brightness = 0;
    int g_display_state = -1;
    int s_fnf5_interface = -1;

    while (( op = getopt( argc, argv, "d:f:hgs:") ) != EOF )
    {
        switch ( op )
        {
            case 'd':
                g_display_state = atoi( optarg );
                break;
            case 'f':
                s_fnf5_interface = atoi( optarg );
                break;
            case 'g':
                g_brightness = 1;
                break;
            case 'h':
                usage( argv[0] );
                exit( 0 );
                break;
            case 's':
                s_brightness = atoi( optarg );
                if ( s_brightness > 7 || s_brightness < 0 )
                {
                    s_brightness = 4;
                }
                break;
            default:
                usage( argv[0] );
                exit( 1 );
                break;
        }
    }

    if ( s_fnf5_interface != -1 )
    {
        set_fnf5_interface( s_fnf5_interface );
    }

    if ( g_display_state != -1 )
    {
        printf( "display state: %d\n", get_display_state() );

        set_display_state( g_display_state );

        printf( "display state: %d\n", get_display_state() );
    }

    if ( g_brightness == 1 )
    {
        printf( "brightness level: %d\n", get_brightness() );
    }

    if ( s_brightness != -1 )
    {
        set_brightness( s_brightness );
        printf( "brightness level: %d\n", get_brightness() );
    }
}
